/* ssa_words_interface.h
 * Copyright (C) 2010, Francisco Claude, all rights reserved.
 *
 * Implementation of the interface "../intIndex/interfaceIntIndex.h"
 *   that permits to represent a sequence of uint32 integers with an iSSA: 
 *   An integer-oriented Compressed Suffix Array.
 *   Such representation will be handled as a "ssa_words" object, that
 *   the WSSA self-index will use (as an opaque type) to 
 *   create/save/load/search/recover/getSize of the representation of the 
 *   original sequence.
 *   
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include<string>
#include<vector>
#include<sstream>
#include<iostream>
#include<algorithm>
#include<ssa_words.h>
#include<interfaceIntIndex.h>
#include<static_sequence_builder.h>
#include<static_bitsequence_builder.h>

using namespace std;

uint touint(string s) {
  stringstream ss;
  ss << s;
  uint ret;
  ss >> ret;
  return ret;
}


void tokenize(string str, vector<string> &tokens, char delim) {
  string::size_type last_pos = 0;
  string::size_type pos = str.find_first_of(delim);
  while(pos!=string::npos) {
    tokens.push_back(str.substr(last_pos,pos-last_pos));
    last_pos = pos+1;
    if(last_pos >= str.length()) break;
    pos = str.find_first_of(delim,pos+1);
  }
  if(last_pos<str.length())
    tokens.push_back(str.substr(last_pos));
}


int buildIntIndex (uint *intVector, uint n, char *build_options, void **index ) {
  uint * myVector = new uint[n+1];
  for(uint i=0;i<n;i++)
    myVector[i] = intVector[i]+1;
  myVector[n] = 0;
  ssa_words * ret = new ssa_words(myVector,n);
  // possample
  uint possample = 64;
  // suffsample
  uint suffsample = 64;
  // bitsample
  uint bitsample1 = 32;
  uint bitsample2 = 32;
  // bitsequence 1
  string sbb1_type = "rrr02";
  // bitsequence 2
  string sbb2_type = "rrr02";
  // sequence
  string ssb_type = "wt_noptr";
  //string huff = "false";
  if(build_options!=NULL) {
    string s = string(build_options);
    cout << "\n build_options SSA_words:<< " << s << ">>\n";

    vector<string> params;
    tokenize(s,params,';');
    for(vector<string>::iterator p=params.begin();p!=params.end();p++) {
      string param = *p;
      vector<string> vals;
      tokenize(param,vals,'=');
      if(vals[0]==string("suffsample"))
        suffsample=touint(vals[1]);
      else if(vals[0]==string("possample"))
        possample=touint(vals[1]);
      else if(vals[0]==string("bitsample1"))
        bitsample1=touint(vals[1]);
      else if(vals[0]==string("bitsample2"))
        bitsample2=touint(vals[1]);
      else if(vals[0]==string("bitseq1_type"))
        sbb1_type = vals[1];
      else if(vals[0]==string("bitseq2_type"))
        sbb2_type = vals[1];
      else if(vals[0]==string("seq_type"))
        ssb_type = vals[1];
      //else if(vals[0]==string("huff"))
      //  huff = vals[1];
    }
  }
  static_bitsequence_builder * sbb = NULL;
  if(sbb1_type==string("rrr02")){
	cout << "Using rrr02 for bitsequence 1";
    sbb = new static_bitsequence_builder_rrr02int(bitsample1);
  }
  else if(sbb1_type==string("brw32")){
	cout << "Using brw32 for bitsequence 1";
    sbb = new static_bitsequence_builder_brw32(bitsample1);
  }
  else {
    cout << "Invalid parameter for sbb1_type (" << sbb1_type << "). ";
    cout << "Using default 'rrr02'" << endl;
    sbb = new static_bitsequence_builder_rrr02int(bitsample1);
  }
  static_bitsequence_builder * sbb2 = NULL;
  if(sbb2_type==string("rrr02")) {
	cout << "Using rrr02 for bitsequence 2";	
    sbb2 = new static_bitsequence_builder_rrr02int(bitsample2);
  }
  else if(sbb2_type==string("brw32")){
	cout << "Using brw32 for bitsequence 2";	
    sbb2 = new static_bitsequence_builder_brw32(bitsample2);
  }
  else {
    cout << "Invalid parameter for sbb2_type (" << sbb2_type << "). ";
    cout << "Using default 'rrr02'" << endl;
    sbb2 = new static_bitsequence_builder_rrr02int(bitsample2);
  }
  static_sequence_builder * ssb = NULL;
  if(ssb_type==string("wt_noptr")) {
	cout << "Using ssb_type =  wavelet_tree_noptrs";
    ssb = new static_sequence_builder_wvtree_noptrs(sbb2, new alphabet_mapper_none());
  }
  else if(ssb_type==string("wt_ptr")) {
	cout << "Using ssb_type = wavelet_tree_ptrs";
    alphabet_mapper *am = new alphabet_mapper_none();
    wt_coder * wc = new wt_coder_huff(myVector,n+1,am);
    ssb = new static_sequence_builder_wvtree(wc,sbb2,am);
  }
  else {
    cout << "Invalid parameter for ssb1_type. ";
    cout << "Using default 'wt_noptr'" << endl;
    ssb = new static_sequence_builder_wvtree_noptrs(sbb2, new alphabet_mapper_none());
  }
  // build
  ret->set_static_bitsequence_builder(sbb);
  ret->set_static_sequence_builder(ssb);
  ret->set_samplepos(possample);
  ret->set_samplesuff(suffsample);
  ret->build_index();
  *index = (void*)ret;
  delete [] myVector;
  return 0;
}


//Saves the index to disk
int saveIntIndex(void *index, char *pathname) {
  FILE * fp = fopen((string(pathname)+".ssa").c_str(),"w");
  if(fp==NULL) return 1;
  ssa_words * ssa = (ssa_words*)index;
  return !ssa->save(fp);
}


//Returns number of elements in the indexed sequence of integers
int sourceLenIntIndex(void *index, uint *numInts) {
  ssa_words * ssa = (ssa_words*)index;
  *numInts = ssa->length();
  return 0;
}


//Loads the index from disk.
int loadIntIndex(char *pathname, void **index) {
  bool err=false;
  FILE * fp = fopen((string(pathname)+".ssa").c_str(),"r");
  if(fp==NULL) return 2;
  ssa_words * ret = new ssa_words(fp,err);
  if(err) return 1;
  *index = (void*)ret;
  fclose(fp);
  return 0;
}


//Frees the memory allocated to the int_index
int freeIntIndex(void *index) {
  ssa_words * ssa = (ssa_words*)index;
  delete ssa;
  return 0;
}


//Returns the size (in bytes) of the index over the sequence of integers.
int sizeIntIndex(void *index, uint *numBytes) {
  ssa_words * ssa = (ssa_words*)index;
  *numBytes = ssa->size();
  return 0;
}


// Shows detailed summary info of the self-index (memory usage of each structure)
int printInfoIntIndex(void *index, const char tab[]) {
  ssa_words * ssa = (ssa_words*)index;
  ssa->print_stats();
  return 0;
}


//Number of occurrences of the pattern, and the interval [left,right] in the suffix array
int countIntIndex(void *index, uint *pattern, uint length, ulong *numocc, ulong *left, ulong *right) {
  //for(uint i=0;i<length;i++) pattern[i]++;
  ssa_words * ssa = (ssa_words*)index;
  *numocc = ssa->count(pattern,length);
  //for(uint i=0;i<length;i++) pattern[i]--;
  return 0;
}


//returns an array with integers corresponding offsets to the occurrences of the pattern,
//as well as the number of occurrences
int locateIntIndex(void *index, uint *pattern, uint length, ulong **occ, ulong *numocc) {
  //for(uint i=0;i<length;i++) pattern[i]++;
  ssa_words * ssa = (ssa_words*)index;
  *numocc = (ulong)ssa->locate(pattern, length, (uint**)occ);
  //for(uint i=0;i<length;i++) pattern[i]--;
  return 0;
}


//Returns the value of the source (array of integers) at a given offset.
// (that is, the element "position" from the original array of uints)
int displayIntIndex(void *index, ulong position, uint *value) {
  ssa_words * ssa = (ssa_words*)index;
  *value = ssa->extract_pos(position);
  return 0;
}
